﻿using System;
using System.Linq;
using System.Threading;
using HeuristicLab.Algorithms.GrammaticalOptimization;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.GrammaticalOptimization;
using HeuristicLab.Selection;
using HeuristicLab.Algorithms.GeneticAlgorithm;

namespace HeuristicLab.Algorithms.GeneticProgramming {
  public class OffspringSelectionGP : SolverBase {
    public int PopulationSize { get; set; }
    public double MutationRate { get; set; }
    public int MaxSolutionSize { get; set; }
    public int MaxSolutionDepth { get; set; }

    private readonly ISymbolicExpressionTreeProblem problem;
    private readonly Random random;

    public OffspringSelectionGP(ISymbolicExpressionTreeProblem problem, Random random) {
      this.problem = problem;
      this.random = random;
      // default parameter values
      PopulationSize = 100;
      MutationRate = 0.15;
      MaxSolutionSize = 100;
      MaxSolutionDepth = 17;
    }

    public override void Run(int maxEvaluations) {
      var hlProblem = new GenericSymbExprProblem(problem);
      var onEvalLocker = new object();
      hlProblem.Evaluator.SolutionEvaluated += (sentence, quality) => {
        // raise solution evaluated event for each GP solution, don't scale quality to 0..1 
        // need to synchronize in case we are using a parallel engine
        lock (onEvalLocker) {
          OnSolutionEvaluated(sentence, quality);
        }
      };
      hlProblem.MaximumSymbolicExpressionTreeLength.Value = MaxSolutionSize;
      hlProblem.MaximumSymbolicExpressionTreeDepth.Value = MaxSolutionDepth;

      using (var wh = new AutoResetEvent(false)) {
        var osga = new OffspringSelectionGeneticAlgorithm.OffspringSelectionGeneticAlgorithm();
        osga.Engine = new ParallelEngine.ParallelEngine();
        osga.ExceptionOccurred += (sender, args) => { Console.WriteLine(args.Value.Message); wh.Set(); };
        osga.Stopped += (sender, args) => { wh.Set(); };

        osga.Problem = hlProblem;
        var mutator = (MultiSymbolicExpressionTreeManipulator)osga.MutatorParameter.ValidValues.Single(op => op.Name == "MultiSymbolicExpressionTreeManipulator");
        foreach (var op in mutator.Operators) {
          if (op.Name == "ChangeNodeTypeManipulation"
            || op.Name == "ReplaceBranchManipulation") mutator.Operators.SetItemCheckedState(op, true);
          else mutator.Operators.SetItemCheckedState(op, false);
        }
        osga.Mutator = mutator;
        osga.Crossover = osga.CrossoverParameter.ValidValues.Single(op => op.Name == "SubtreeSwappingCrossover");
        osga.Selector = osga.SelectorParameter.ValidValues.Single(op => op.Name == "GenderSpecificSelection");

        osga.PopulationSize.Value = PopulationSize;
        osga.MaximumGenerations.Value = 100000; // some very large value (we stop based on evaluations)
        osga.MaximumSelectionPressure.Value = 1000;
        osga.MaximumEvaluatedSolutions.Value = maxEvaluations;
        osga.MutationProbability.Value = MutationRate;

        osga.SetSeedRandomly = new BoolValue(false);
        osga.Seed = new IntValue(random.Next());

        osga.Prepare();
        osga.Start();

        wh.WaitOne();
      }
    }
  }
}
